@@ -9,7 +9,7 @@ static PangoWeight weight[] = { PANGO_WEIGHT_NORMAL, PANGO_WEIGHT_BOLD };
 CairoDesc *createCairoDesc() {
   return(g_new0(CairoDesc, 1));
 }
-void freeCairoDesc(NewDevDesc *dd) {
+void freeCairoDesc(pDevDesc dd) {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   
   if (!cd)
@@ -39,18 +39,16 @@ void freeCairoDesc(NewDevDesc *dd) {
   g_free(cd);
 }
 
-static void activateDevice(NewDevDesc *dev)
+static void activateDevice(pDevDesc dev)
 {
-  GEDevDesc *dd;
+  pGEDevDesc gdd;
   CairoDesc *cd = (CairoDesc *)dev->deviceSpecific;
   GObject *obj = cd->drawing ? G_OBJECT(cd->drawing) : G_OBJECT(cd->pixmap);
-  gsetVar(install(".Device"), mkString("Cairo"), R_NilValue);
-  dd = GEcreateDevDesc(dev);
-  /*dd->newDevStruct = 1;*/
-  Rf_addDevice((DevDesc*) dd);
-  GEinitDisplayList(dd);
+
+  gdd = GEcreateDevDesc(dev);
+  GEaddDevice2(gdd, "Cairo");
   if (obj) {
-    SEXP devnum = ScalarInteger(Rf_devNumber((DevDesc*)dev) + 1);
+    SEXP devnum = ScalarInteger(ndevNumber(dev) + 1);
     R_PreserveObject(devnum);
     g_object_set_data_full(obj, ".devnum", devnum, (GDestroyNotify)R_ReleaseObject);
   }
@@ -90,7 +88,7 @@ static PangoFontDescription *getBaseFont(CairoDesc *cd)
   return(pango_font_description_from_string("Verdana"));
 }
 
-static void blank(NewDevDesc *dd) {
+static void blank(pDevDesc dd) {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   cairo_t *cr = cd->cr;
   gint width = cd->width, height = cd->height;
@@ -104,7 +102,7 @@ static void blank(NewDevDesc *dd) {
     gtk_widget_queue_draw(cd->drawing);
 }
 
-static gboolean initDevice(NewDevDesc *dd)
+static gboolean initDevice(pDevDesc dd)
 {
   CairoDesc *cd;
   GdkCursor *cursor;
@@ -151,18 +149,18 @@ static gboolean initDevice(NewDevDesc *dd)
   return FALSE;
 }
 
-static void resize(NewDevDesc *dd)
+static void resize(pDevDesc dd)
 {
-  GEplayDisplayList ((GEDevDesc*) Rf_GetDevice(Rf_devNumber((DevDesc*)dd)));
+    GEplayDisplayList(desc2GEDesc(dd));
 }
 
-static gboolean realize_event(GtkWidget *widget, NewDevDesc *dd)
+static gboolean realize_event(GtkWidget *widget, pDevDesc dd)
 {
   g_return_val_if_fail(dd != NULL, FALSE);
   initDevice(dd);
   return(FALSE);
 }
-static gboolean realize_embedded(GtkWidget *widget, NewDevDesc *dd)
+static gboolean realize_embedded(GtkWidget *widget, pDevDesc dd)
 { /* different from above in that embedded device needs to be activated,
      since it wasn't when the device was created */
   g_return_val_if_fail(dd != NULL, FALSE);
@@ -171,7 +169,7 @@ static gboolean realize_embedded(GtkWidget *widget, NewDevDesc *dd)
   return(FALSE);
 }
 
-static gint expose_event(GtkWidget *widget, GdkEventExpose *event, NewDevDesc *dd)
+static gint expose_event(GtkWidget *widget, GdkEventExpose *event, pDevDesc dd)
 {
   CairoDesc *cd;
     
@@ -195,7 +193,7 @@ static gint expose_event(GtkWidget *widget, GdkEventExpose *event, NewDevDesc *d
   return FALSE;
 }
 
-static void event_finish(NewDevDesc *dd)
+static void event_finish(pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   dd->onExit = NULL;
@@ -203,13 +201,13 @@ static void event_finish(NewDevDesc *dd)
   cd->event->active = FALSE;
   cd->event = NULL;
 }
-static void event_maybe_finish(NewDevDesc *dd)
+static void event_maybe_finish(pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (cd->event->result && cd->event->result != R_NilValue)
     event_finish(dd);
 }
-static void CairoEvent_onExit(NewDevDesc *dd)
+static void CairoEvent_onExit(pDevDesc dd)
 {
   event_finish(dd);
 }
@@ -218,7 +216,7 @@ static void CairoEvent_onExit(NewDevDesc *dd)
 #define R_BUTTON(button) pow(2, (button) - 1)
 #define R_BUTTONS(state) (state) & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK | GDK_BUTTON3_MASK) >> 8
 
-static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event, NewDevDesc *dd)
+static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
@@ -229,7 +227,7 @@ static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event, New
   return(FALSE);
 }
 
-static gboolean button_release_event(GtkWidget *widget, GdkEventButton *event, NewDevDesc *dd)
+static gboolean button_release_event(GtkWidget *widget, GdkEventButton *event, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
@@ -240,7 +238,7 @@ static gboolean button_release_event(GtkWidget *widget, GdkEventButton *event, N
   return(FALSE);
 }
 
-static gboolean motion_notify_event(GtkWidget *widget, GdkEventMotion *event, NewDevDesc *dd)
+static gboolean motion_notify_event(GtkWidget *widget, GdkEventMotion *event, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
@@ -251,7 +249,7 @@ static gboolean motion_notify_event(GtkWidget *widget, GdkEventMotion *event, Ne
   return(FALSE);
 }
 
-static gboolean key_press_event(GtkWidget *widget, GdkEventKey *event, NewDevDesc *dd)
+static gboolean key_press_event(GtkWidget *widget, GdkEventKey *event, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
@@ -261,9 +259,9 @@ static gboolean key_press_event(GtkWidget *widget, GdkEventKey *event, NewDevDes
   return(FALSE);
 }
 
-static SEXP Cairo_GetEvent(SEXP rho, char *prompt)
+static SEXP Cairo_GetEvent(SEXP rho, const char *prompt)
 {
-  GEDevDesc *dd = GEcurrentDevice();
+  pGEDevDesc dd = GEcurrentDevice();
   CairoDesc *cd = (CairoDesc *) dd->dev->deviceSpecific;
   CairoEvent *event = g_new0(CairoEvent, 1);
   SEXP result = R_NilValue;
@@ -293,12 +291,12 @@ static SEXP Cairo_GetEvent(SEXP rho, char *prompt)
   return result;
 }
 
-static void kill_cairo(NewDevDesc *dd)
+static void kill_cairo(pDevDesc dd)
 {
-  Rf_KillDevice((DevDesc*) Rf_GetDevice(Rf_devNumber ((DevDesc*) dd)));
+    GEkillDevice(desc2GEDesc(dd));
 }
 
-static void unrealize_cb(GtkWidget *widget, NewDevDesc *dd) {
+static void unrealize_cb(GtkWidget *widget, pDevDesc dd) {
   g_return_if_fail(dd != NULL);
   if (dd->deviceSpecific) { /* make sure not called from freeCairoDesc */
     /* don't try to destroy the widget, since it's already being destroyed */
@@ -306,16 +304,15 @@ static void unrealize_cb(GtkWidget *widget, NewDevDesc *dd) {
     kill_cairo(dd);
   }
 }
-static gint delete_event(GtkWidget *widget, GdkEvent *event, NewDevDesc *dd)
+static gint delete_event(GtkWidget *widget, GdkEvent *event, pDevDesc dd)
 {
   g_return_val_if_fail(dd != NULL, FALSE);
   kill_cairo(dd);
   return TRUE;
 }
 
-static void setupWidget(GtkWidget *drawing, NewDevDesc *dd)
+static void setupWidget(GtkWidget *drawing, pDevDesc dd)
 {
-  GtkWidget *wid = drawing;
   gtk_widget_add_events(drawing, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
                         GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);
   g_signal_connect(G_OBJECT(drawing), "expose_event", G_CALLBACK(expose_event), dd);
@@ -327,7 +324,7 @@ static void setupWidget(GtkWidget *drawing, NewDevDesc *dd)
                    G_CALLBACK(button_press_event), dd);
 }
 
-static Rboolean Cairo_OpenEmbedded(NewDevDesc *dd, CairoDesc *cd, GtkWidget *drawing)
+static Rboolean Cairo_OpenEmbedded(pDevDesc dd, CairoDesc *cd, GtkWidget *drawing)
 {
   dd->deviceSpecific = cd;
   cd->drawing = drawing;
@@ -341,7 +338,7 @@ static Rboolean Cairo_OpenEmbedded(NewDevDesc *dd, CairoDesc *cd, GtkWidget *dra
   g_signal_connect(G_OBJECT(drawing), "unrealize", G_CALLBACK(unrealize_cb), dd);
   return(TRUE);
 }
-static Rboolean Cairo_OpenOffscreen(NewDevDesc *dd, CairoDesc *cd, GdkDrawable *drawing)
+static Rboolean Cairo_OpenOffscreen(pDevDesc dd, CairoDesc *cd, GdkDrawable *drawing)
 {
   dd->deviceSpecific = cd;
   cd->pixmap = drawing;
@@ -350,7 +347,7 @@ static Rboolean Cairo_OpenOffscreen(NewDevDesc *dd, CairoDesc *cd, GdkDrawable *
   initDevice(dd);		 
   return(TRUE);
 }
-static Rboolean Cairo_Open(NewDevDesc *dd, CairoDesc *cd,	double w, double h, 
+static Rboolean Cairo_Open(pDevDesc dd, CairoDesc *cd,	double w, double h, 
                            const gchar **surface_info)
 {	
   dd->deviceSpecific = cd;
@@ -419,7 +416,7 @@ static Rboolean Cairo_Open(NewDevDesc *dd, CairoDesc *cd,	double w, double h,
   return(TRUE);
 }
 
-static PangoFontDescription *getFont(CairoDesc *cd, R_GE_gcontext *gc)
+static PangoFontDescription *getFont(CairoDesc *cd, const pGEcontext gc)
 {
   PangoFontDescription *fontdesc, *basedesc;
   gint size, face = gc->fontface;
@@ -453,10 +450,9 @@ static PangoFontDescription *getFont(CairoDesc *cd, R_GE_gcontext *gc)
   return(fontdesc);
 }
 
-static PangoLayout *layoutText(PangoFontDescription *desc, const char *str, CairoDesc *cd)
+static PangoLayout *layoutText(PangoFontDescription *desc, const char *str, 
+			       CairoDesc *cd)
 {
-  gsize bytes_written;
-  gchar *utf8;
   PangoLayout *layout;
 	
   //pango_cairo_update_context(cd->cr, cd->pango);
@@ -465,84 +461,21 @@ static PangoLayout *layoutText(PangoFontDescription *desc, const char *str, Cair
     layout = gtk_widget_create_pango_layout(cd->drawing, NULL);
   else layout = pango_layout_new(gdk_pango_context_get());
   pango_layout_set_font_description(layout, desc);
-  utf8 = g_locale_to_utf8(str, -1, NULL, &bytes_written, NULL);
-  pango_layout_set_text(layout, utf8, -1);
-	
-  g_free(utf8);
-	
-  return(layout);
-}
-
-/* Adobe -> UTF8 translation courtesy of Mikhail Kondrin */
-static PangoLayout *layoutMath(PangoFontDescription *desc, const char *str, CairoDesc *cd)
-{
-  gsize bytes_written;
-  gchar *utf8;
-  PangoLayout *layout;
-  GString * z = g_string_new(NULL);
-  gunichar unisym[190]={
-    0x0020, 0x0021, 0x2200, 0x0023, 0x2203, 0x0025, 0x0026, 0x220B, 0x0028,
-    0x0029, 0x2217, 0x002B, 0x002C, 0x2212, 0x002E, 0x002F, 0x0030, 0x0031,
-    0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 0x0038, 0x0039, 0x003A,
-    0x003B, 0x003C, 0x003D, 0x003E, 0x003F, 0x2245, 0x0391, 0x0392, 0x03A7,
-    0x0394, 0x0395, 0x03A6, 0x0393, 0x0397, 0x0399, 0x03D1, 0x039A, 0x039B,
-    0x039C, 0x039D, 0x039F, 0x03A0, 0x0398, 0x03A1, 0x03A3, 0x03A4, 0x03A5,
-    0x03C2, 0x2126, 0x039E, 0x03A8, 0x0396, 0x005B, 0x2234, 0x005D, 0x22A5,
-    0x005F, 0xF8E5, 0x03B1, 0x03B2, 0x03C7, 0x03B4, 0x03B5, 0x03C6, 0x03B3,
-    0x03B7, 0x03B9, 0x03D5, 0x03BA, 0x03BB, 0x03BC, 0x03BD, 0x03BF, 0x03C0,
-    0x03B8, 0x03C1, 0x03C3, 0x03C4, 0x03C5, 0x03D6, 0x03C9, 0x03BE, 0x03C8,
-    0x03B6, 0x007B, 0x007C, 0x007D, 0x223C, 0x20AC, 0x03D2, 0x2032, 0x2264,
-    0x2044, 0x221E, 0x0192, 0x2663, 0x2666, 0x2665, 0x2660, 0x2194, 0x2190,
-    0x2191, 0x2192, 0x2193, 0x00B0, 0x00B1, 0x2033, 0x2265, 0x00D7, 0x221D,
-    0x2202, 0x2022, 0x00F7, 0x2260, 0x2261, 0x2248, 0x2026, 0xF8E6, 0xF8E7,
-    0x21B5, 0x2135, 0x2111, 0x211C, 0x2118, 0x2297, 0x2295, 0x2205, 0x2229,
-    0x222A, 0x2283, 0x2287, 0x2284, 0x2282, 0x2286, 0x2208, 0x2209, 0x2220,
-    0x2207, 0xF6DA, 0xF6D9, 0xF6DB, 0x220F, 0x221A, 0x22C5, 0x00AC, 0x2227,
-    0x2228, 0x21D4, 0x21D0, 0x21D1, 0x21D2, 0x21D3, 0x25CA, 0x2329, 0xF8E8,
-    0xF8E9, 0xF8EA, 0x2211, 0xF8EB, 0xF8EC, 0xF8ED, 0xF8EE, 0xF8EF, 0xF8F0,
-    0xF8F1, 0xF8F2, 0xF8F3, 0xF8F4, 0x0020, 0x232A, 0x222B, 0x2320, 0xF8F5,
-    0x2321, 0xF8F6, 0xF8F7, 0xF8F8, 0xF8F9, 0xF8FA, 0xF8FB, 0xF8FC, 0xF8FD,
-    0xF8FE
-  };
-  gint i;
-  guchar j;
-  gunichar z0;
-  
-  if (cd->drawing)
-    layout = gtk_widget_create_pango_layout(cd->drawing, NULL);
-  else layout = pango_layout_new(gdk_pango_context_get());
-  pango_layout_set_font_description(layout, desc);
-  
-  z0=0x20;
-  for (i=0;i<strlen(str);i++){
-    j=*(str+i);
-    if ((j>=0x20) & (j<=0x7F))
-      z0=unisym[j-32];
-    else if ((j>=0xA0) & (j<=0xFE))
-      z0=unisym[j-65];
-    g_string_append_unichar(z,z0);
-  }
-  
-  pango_layout_set_text(layout, z->str, -1);
-  
-  g_string_free(z,1);
-  
+  pango_layout_set_text(layout, str, -1);
   return(layout);
 }
 
 static void
-text_extents(PangoFontDescription *desc, CairoDesc *cd, R_GE_gcontext *gc, 
-             const gchar *text, gint *lbearing, gint *rbearing, gint *width, gint *ascent, gint *descent)
+text_extents(PangoFontDescription *desc, CairoDesc *cd, const pGEcontext gc, 
+             const gchar *text,
+	     gint *lbearing, gint *rbearing, 
+	     gint *width, gint *ascent, gint *descent)
 {
   PangoLayout *layout;
   PangoRectangle rect;
 	
-  if (gc->fontface!=SYMBOL_FONTFACE) {
-    layout = layoutText(desc, text, cd);
-  } else {
-    layout = layoutMath(desc, text, cd);
-  }
-	
+  layout = layoutText(desc, text, cd);
+  
   pango_layout_line_get_pixel_extents(pango_layout_get_line(layout, 0), NULL, &rect);
 
   if(ascent)
@@ -580,7 +513,7 @@ static void setColor(cairo_t *cr, int color)
 }
 
 /* set the line type */
-static void setLineType(cairo_t *cr, R_GE_gcontext *gc)
+static void setLineType(cairo_t *cr, const pGEcontext gc)
 {
   cairo_line_cap_t cap = CAIRO_LINE_CAP_ROUND;
   cairo_line_join_t join = CAIRO_LINE_JOIN_ROUND;
@@ -628,7 +561,7 @@ static void setLineType(cairo_t *cr, R_GE_gcontext *gc)
   cairo_set_dash(cr, dashes, i, 0);
 }
 
-static void drawShape(cairo_t *cr, R_GE_gcontext *gc)
+static void drawShape(cairo_t *cr, const pGEcontext gc)
 {
   if (gc->fill != NA_INTEGER) {
     setColor(cr, gc->fill);
@@ -642,13 +575,13 @@ static void drawShape(cairo_t *cr, R_GE_gcontext *gc)
 }
 
 static void drawRect(cairo_t *cr, double x0, double y0, double x1, double y1, 
-                     R_GE_gcontext *gc) 
+                     const pGEcontext gc) 
 {
   cairo_rectangle(cr, x0, y0, x1 - x0, y1 - y0);
   drawShape(cr, gc);
 }
 
-static double Cairo_StrWidth(char *str, R_GE_gcontext *gc, NewDevDesc *dd)
+static double Cairo_StrWidth(const char *str, const pGEcontext gc, pDevDesc dd)
 {
   gint width;
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
@@ -660,20 +593,26 @@ static double Cairo_StrWidth(char *str, R_GE_gcontext *gc, NewDevDesc *dd)
   return (double) width;
 }
 
-static void Cairo_MetricInfo(int c, R_GE_gcontext *gc,
-                             double* ascent, double* descent, double* width, NewDevDesc *dd)
+static void Cairo_MetricInfo(int c, const pGEcontext gc,
+                             double* ascent, double* descent, double* width, pDevDesc dd)
 {
-  gchar text[2];
+  gchar text[16];
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   gint iascent, idescent, iwidth;
+  int Unicode = mbcslocale;
 	
   PangoFontDescription *desc = getFont(cd, gc);
 	
   if (!c) 
     font_metrics(desc, cd, &iwidth, &iascent, &idescent);
   else {
-    g_snprintf(text, 2, "%c", (gchar) c);
-    text_extents(desc, cd, gc, text, NULL, NULL, &iwidth, &iascent, &idescent);
+      if(c < 0) {c = -c; Unicode = 1;} 
+      
+      if(Unicode || c >= 128)
+	  Rf_ucstoutf8(text, c);
+      else
+	  g_snprintf(text, 2, "%c", (gchar) c);
+      text_extents(desc, cd, gc, text, NULL, NULL, &iwidth, &iascent, &idescent);
   }
 	
   *ascent = iascent;
@@ -685,7 +624,7 @@ static void Cairo_MetricInfo(int c, R_GE_gcontext *gc,
   pango_font_description_free(desc);
 }
 
-static void Cairo_Clip(double x0, double x1, double y0, double y1, NewDevDesc *dd)
+static void Cairo_Clip(double x0, double x1, double y0, double y1, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   gint cx, cy, cw, ch;
@@ -705,7 +644,7 @@ static void Cairo_Clip(double x0, double x1, double y0, double y1, NewDevDesc *d
 }
 
 static void Cairo_Size(double *left, double *right, double *bottom, double *top,
-                       NewDevDesc *dd)
+                       pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   gint width = cd->width, height = cd->height;
@@ -727,7 +666,7 @@ static void Cairo_Size(double *left, double *right, double *bottom, double *top,
 }
 
 /* clear the drawing area */
-static void Cairo_NewPage(R_GE_gcontext *gc, NewDevDesc *dd)
+static void Cairo_NewPage(const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = dd->deviceSpecific;
   gint width = cd->width, height = cd->height;
@@ -746,7 +685,7 @@ static void Cairo_NewPage(R_GE_gcontext *gc, NewDevDesc *dd)
 
 /** kill off the window etc
  */
-static void Cairo_Close(NewDevDesc *dd)
+static void Cairo_Close(pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->onExit)
@@ -760,7 +699,7 @@ static void Cairo_Close(NewDevDesc *dd)
 #define title_text_active "R graphics device %d - Active"
 
 void
-setActive(NewDevDesc *dd, gboolean active)
+setActive(pDevDesc dd, gboolean active)
 {
   CairoDesc *cd;
   gint devnum;
@@ -772,7 +711,7 @@ setActive(NewDevDesc *dd, gboolean active)
   if (!cd->window)
     return;
 
-  devnum = Rf_devNumber((DevDesc*)dd) + 1;
+  devnum = ndevNumber(dd) + 1;
 
   if (active)
     title_text = g_strdup_printf(title_text_active, devnum);
@@ -783,18 +722,18 @@ setActive(NewDevDesc *dd, gboolean active)
   g_free(title_text);
 }
 
-static void Cairo_Activate(NewDevDesc *dd)
+static void Cairo_Activate(pDevDesc dd)
 {
   setActive(dd, TRUE);
 }
 
-static void Cairo_Deactivate(NewDevDesc *dd)
+static void Cairo_Deactivate(pDevDesc dd)
 {
   setActive(dd, FALSE);
 }	
 
 static void Cairo_Rect(double x0, double y0, double x1, double y1,
-                       R_GE_gcontext *gc, NewDevDesc *dd)
+                       const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
 	
@@ -806,7 +745,7 @@ static void Cairo_Rect(double x0, double y0, double x1, double y1,
   cairo_restore(cd->cr);
 }
 
-static void drawCircle(cairo_t *cr, double x, double y, double r, R_GE_gcontext *gc)
+static void drawCircle(cairo_t *cr, double x, double y, double r, const pGEcontext gc)
 {
   cairo_move_to(cr, x+r, y);
   cairo_translate(cr, x, y);
@@ -816,7 +755,7 @@ static void drawCircle(cairo_t *cr, double x, double y, double r, R_GE_gcontext
 }
 
 static void Cairo_Circle(double x, double y, double r,
-                         R_GE_gcontext *gc, NewDevDesc *dd)
+                         const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   
@@ -829,7 +768,7 @@ static void Cairo_Circle(double x, double y, double r,
 }
 
 static void drawLine(cairo_t *cr, double x1, double y1, double x2, double y2,
-                     R_GE_gcontext *gc)
+                     const pGEcontext gc)
 {
   cairo_move_to(cr, x1, y1);
   cairo_line_to(cr, x2, y2);
@@ -840,7 +779,7 @@ static void drawLine(cairo_t *cr, double x1, double y1, double x2, double y2,
 }
 
 static void Cairo_Line(double x1, double y1, double x2, double y2,
-                       R_GE_gcontext *gc, NewDevDesc *dd)
+                       const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   //g_debug("line");
@@ -861,7 +800,7 @@ static void polypath(cairo_t *cr, int n, double *x, double *y)
     cairo_line_to(cr, x[i], y[i]);
 }
 
-static void drawPolyline(cairo_t *cr, int n, double *x, double *y, R_GE_gcontext *gc)
+static void drawPolyline(cairo_t *cr, int n, double *x, double *y, const pGEcontext gc)
 {
   polypath(cr, n, x, y);
   setColor(cr, gc->col);
@@ -870,7 +809,7 @@ static void drawPolyline(cairo_t *cr, int n, double *x, double *y, R_GE_gcontext
 }
 
 static void Cairo_Polyline(int n, double *x, double *y, 
-                           R_GE_gcontext *gc, NewDevDesc *dd)
+                           const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   //g_debug("polyline");
@@ -882,7 +821,7 @@ static void Cairo_Polyline(int n, double *x, double *y,
   cairo_restore(cd->cr);
 }
 
-static void drawPolygon(cairo_t *cr, int n, double *x, double *y, R_GE_gcontext *gc)
+static void drawPolygon(cairo_t *cr, int n, double *x, double *y, const pGEcontext gc)
 {
   polypath(cr, n, x, y);
   cairo_close_path(cr);
@@ -890,7 +829,7 @@ static void drawPolygon(cairo_t *cr, int n, double *x, double *y, R_GE_gcontext
 }
 
 static void Cairo_Polygon(int n, double *x, double *y, 
-                          R_GE_gcontext *gc, NewDevDesc *dd)
+                          const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
 	
@@ -902,8 +841,8 @@ static void Cairo_Polygon(int n, double *x, double *y,
   cairo_restore(cd->cr);
 }
 
-static void drawText(double x, double y, char *str, 
-		     double rot, double hadj, CairoDesc *cd, R_GE_gcontext *gc)
+static void drawText(double x, double y, const char *str, 
+		     double rot, double hadj, CairoDesc *cd, const pGEcontext gc)
 {
   PangoLayout *layout;
   gint ascent, lbearing;
@@ -919,19 +858,15 @@ static void drawText(double x, double y, char *str,
   cairo_rel_move_to(cr, -lbearing, -ascent);
   setColor(cr, gc->col);
 	
-  if (gc->fontface!=SYMBOL_FONTFACE) {
-    layout = layoutText(desc, str, cd);
-  } else {
-    layout = layoutMath(desc, str, cd);
-  }
+  layout = layoutText(desc, str, cd);
   pango_cairo_show_layout(cr, layout);
 	
   g_object_unref(layout);
   pango_font_description_free(desc);
 }
 
-static void Cairo_Text(double x, double y, char *str, 
-                       double rot, double hadj, R_GE_gcontext *gc, NewDevDesc *dd)
+static void Cairo_Text(double x, double y, const char *str, 
+                       double rot, double hadj, const pGEcontext gc, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   double rrot = DEG2RAD * rot;
@@ -943,7 +878,7 @@ static void Cairo_Text(double x, double y, char *str,
   cairo_restore(cd->cr);
 }
 
-static void locator_finish(NewDevDesc *dd)
+static void locator_finish(pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   /* clean up handler */
@@ -955,7 +890,7 @@ static void locator_finish(NewDevDesc *dd)
 
 static void locator_button_press(GtkWidget *widget,
 				 GdkEventButton *event,
-				 NewDevDesc *dd)
+				 pDevDesc dd)
 {
   CairoLocator *info = ((CairoDesc *) dd->deviceSpecific)->locator;
 
@@ -969,12 +904,12 @@ static void locator_button_press(GtkWidget *widget,
   locator_finish(dd);
 }
 
-static void CairoLocator_onExit(NewDevDesc *dd)
+static void CairoLocator_onExit(pDevDesc dd)
 {
   locator_finish(dd);
 }
 
-static Rboolean Cairo_Locator(double *x, double *y, NewDevDesc *dd)
+static Rboolean Cairo_Locator(double *x, double *y, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   CairoLocator *info;
@@ -1019,7 +954,7 @@ static Rboolean Cairo_Locator(double *x, double *y, NewDevDesc *dd)
   return FALSE;
 }
 
-static void Cairo_Mode(gint mode, NewDevDesc *dd)
+static void Cairo_Mode(gint mode, pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
 	
@@ -1031,12 +966,8 @@ static void Cairo_Mode(gint mode, NewDevDesc *dd)
   }
 }
 
-static void Cairo_Hold(NewDevDesc *dd)
-{
-}
-
 Rboolean
-configureCairoDevice(NewDevDesc *dd, CairoDesc *cd, double width, double height, double ps)
+configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, double ps)
 {
   gint ascent, descent, cw;
   PangoFont *success;
@@ -1046,15 +977,13 @@ configureCairoDevice(NewDevDesc *dd, CairoDesc *cd, double width, double height,
     
   dd->deviceSpecific = (void *) cd;
 
-  dd->newDevStruct = 1;
-
-  dd->open = Cairo_Open;
   dd->close = Cairo_Close;
   dd->activate = Cairo_Activate;
   dd->deactivate = Cairo_Deactivate;
   dd->size = Cairo_Size;
   dd->newPage = Cairo_NewPage;
   dd->clip = Cairo_Clip;
+  /* Next two are unused */
   dd->strWidth = Cairo_StrWidth;
   dd->text = Cairo_Text;
   dd->rect = Cairo_Rect;
@@ -1064,9 +993,12 @@ configureCairoDevice(NewDevDesc *dd, CairoDesc *cd, double width, double height,
   dd->polygon = Cairo_Polygon;
   dd->locator = Cairo_Locator;
   dd->mode = Cairo_Mode;
-  dd->hold = Cairo_Hold;
   dd->metricInfo = Cairo_MetricInfo;
   dd->getEvent = Cairo_GetEvent;
+  dd->hasTextUTF8 = TRUE;
+  dd->wantSymbolUTF8 = TRUE;
+  dd->strWidthUTF8 = Cairo_StrWidth;
+  dd->textUTF8 = Cairo_Text;
 
   dd->left = 0;
   dd->right = width;
@@ -1109,10 +1041,6 @@ configureCairoDevice(NewDevDesc *dd, CairoDesc *cd, double width, double height,
   dd->ipr[1] = pixelHeight();
 
   /* device capabilities */
-  dd->canResizePlot = TRUE;
-  dd->canChangeFont = TRUE;
-  dd->canRotateText = TRUE;
-  dd->canResizeText = TRUE;
   dd->canClip = TRUE;
   dd->canHAdj = 0; // maybe 1 or 2? 
   dd->canChangeGamma = FALSE; // not yet
@@ -1132,7 +1060,7 @@ configureCairoDevice(NewDevDesc *dd, CairoDesc *cd, double width, double height,
 }
 
 Rboolean
-createCairoDevice(NewDevDesc *dd, double width, double height, double ps, void *data) {
+createCairoDevice(pDevDesc dd, double width, double height, double ps, void *data) {
   /* device driver start */
 	
   CairoDesc *cd;
@@ -1149,7 +1077,7 @@ createCairoDevice(NewDevDesc *dd, double width, double height, double ps, void *
   return(configureCairoDevice(dd, cd, width / pixelWidth(), height / pixelHeight(), ps));
 }
 Rboolean
-asCairoDevice(NewDevDesc *dd, double width, double height, double ps, void *data)
+asCairoDevice(pDevDesc dd, double width, double height, double ps, void *data)
 {
   CairoDesc *cd;
   gdouble left, right, bottom, top;
@@ -1172,22 +1100,21 @@ asCairoDevice(NewDevDesc *dd, double width, double height, double ps, void *data
   return(configureCairoDevice(dd, cd, right, bottom, ps));
 }
 
-typedef Rboolean (*CairoDeviceCreateFun)(NewDevDesc *, double width, 
+typedef Rboolean (*CairoDeviceCreateFun)(pDevDesc , double width, 
                                          double height, double pointsize, void *data);
 
-static  NewDevDesc *
+static  pDevDesc 
 initCairoDevice(double width, double height, double ps, void *data, CairoDeviceCreateFun init_fun)
 {
-  NewDevDesc *dev;
+  pDevDesc dev;
   CairoDesc *cd;
 
+  R_GE_checkVersionOrDie(R_GE_version);
   R_CheckDeviceAvailable();
   BEGIN_SUSPEND_INTERRUPTS {
     /* Allocate and initialize the device driver data */
-    if (!(dev = (NewDevDesc *) calloc(1, sizeof(NewDevDesc))))
+    if (!(dev = (pDevDesc) calloc(1, sizeof(NewDevDesc))))
       return NULL;
-    /* Do this for early redraw attempts */
-    dev->displayList = R_NilValue;
     if (! init_fun (dev, width, height, ps, data)) {
       free(dev);
       PROBLEM  "unable to start device cairo" ERROR;
